#ifndef _RAY_H_
#define _RAY_H_

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

// Memory address difinitions
#define MGT_SELFNODEID	0x8000a000
#define MGT_ALLNODEID	0x8000a004

#define MASTER_NTIME		0x20001000	// unsinged int
#define MASTER_LINES_X		0x20001004	// unsinged int
#define MASTER_LINES_X_PREV	0x20001324	// unsinged int
#define MASTER_LINES_NODE	0x20001644	// unsigned int
#define MASTER_LINES_POINT	0x20001964	// class Point 
#define MASTER_SLAVE_ADDR	0x2000a000

#define lines_node(y) (((unsigned int *) (MASTER_LINES_NODE)))[y]

#define SLAVE_LINES_X		0	// unsigned int
#define SLAVE_LINE_Y		4	// unsigned int
#define SLAVE_LINES_POINT	8	// class Point

#define slave_retrieve_lines_x(nodeid) \
	* (unsigned int *) (0x2000b000 + (nodeid*0x1000))
#define slave_refer_to_lines_x(nodeid) \
	(unsigned int *) (0x2000b000 + (nodeid*0x1000))
#define slave_retrieve_line_y(nodeid) \
	* (unsigned int *) (0x2000b000 + (nodeid*0x1000) + 4)
#define slave_refer_to_line_y(nodeid) \
	(unsigned int *) (0x2000b000 + (nodeid*0x1000) + 4)
#define slave_retrieve_points(nodeid) \
	(unsigned char *) (0x2000b000 + (nodeid*0x1000) + 8)

#define MIRROR_STARTADDR	0x8000a010
#define MIRROR_ENDADDR		0x8000a014
#define MIRROR_INTERVALADDR	0x8000a018
#define MIRROR_SENDGAPADDR	0x8000a01c

#define DEFAULT_MIRROR_SENDGAP	0

#define False	0
#define True	1

unsigned int	self_nodeid;

unsigned int
get_master_nodeid ()
{
	int	n = 1;
	unsigned int allnode_bits =	* ((unsigned int *) MGT_ALLNODEID);

	for (int i = 1; i <= 32; i++) {
		if (allnode_bits & n)
			return i;
		n <<= 1;
	}

	return 0;
}

void CommitChanges (void *sa, int size)
{
	unsigned int	*sp = (unsigned int *) MIRROR_STARTADDR,
			*ep = (unsigned int *) MIRROR_ENDADDR,
			*gp = (unsigned int *) MIRROR_SENDGAPADDR;
	char	*cp;

	cp = (char *) sa;
	*sp = (unsigned int) sa;
	*ep = (unsigned int) (cp + size);
//printf ("CommitChanges: sp: %p ep: %p\n", sp, ep);
	*gp = DEFAULT_MIRROR_SENDGAP;

	return ;
}

int
is_master(unsigned int nid)
{
	if (get_master_nodeid () == nid) {
		node_type = NODETYPE_MASTER;
		return True;
	} else {
		if (node_type == NODETYPE_MASTER) {
			CommitChanges ((void *) MASTER_NTIME, 34404);
			sleep (2);
		}
		node_type = NODETYPE_SLAVE;
		return False;
	}
}


int
isCalcComplete (unsigned int *lines_x)
{
	for (int i = 0; i < 200; i++) {
printf ("lines_x[%d] : %d\n", i, lines_x[i]);
		if (! lines_x[i]) {
printf ("line %d still not completed \n", i);
			return False;
		}
	}
	return True;
}

void UpdateLinesData(int y) 
{
	unsigned int	node;
	unsigned int	*nodes = (unsigned int *) MASTER_LINES_NODE;
	unsigned int  *xs = (unsigned int *) MASTER_LINES_X;
	unsigned char	*sp, *mp;

//printf ("updatelines %d\n", y);
	node = nodes[y];

//printf ("node %d doing %d x %d\n", node, y, slave_retrieve_lines_x(node));
	mp = (unsigned char *) MASTER_LINES_POINT;
	if (slave_retrieve_lines_x(node) == 320) {
		sp = slave_retrieve_points(node);
		memcpy (&mp[160*y], sp, 160);
//		CommitChanges (&mp[160*y], 160);
		xs[y] = 320;
//		CommitChanges (&xs[y], sizeof (unsigned int));
		CommitChanges ((void *) MASTER_NTIME, 34400);
//printf ("UpdateLIne: line %d completed %d\n", y, xs[y]);
	}
}

void
MasterCheckLinesNode (int y)
{
	unsigned int	*a;
	unsigned int  *xs = (unsigned int *) MASTER_LINES_X;
	unsigned long	allnode_bits = * ((unsigned int *) MGT_ALLNODEID);
	unsigned int	*nodes = (unsigned int *) MASTER_LINES_NODE;

	for (int i = 0; i < 200; i++) {
		int	n = 1;
		int	a = n << (nodes[i]-1);

		if (nodes[i] && (allnode_bits & a)) {
			continue;
		} else if (nodes[i])
			nodes[i] = 0;

		for (int j = 1; j <= 32; j++) {
			a = n << (j-1);
			if (! (allnode_bits & a)) {
				continue;
			}

			if (slave_retrieve_line_y(j) == i) {
				nodes[i] = j;
//				CommitChanges (&nodes[i], 4);
				CommitChanges ((void *) MASTER_NTIME, 34400);
			}
		}
	}

	return ;
}

void
MasterCheckLinesDone (int y) 
{
	unsigned long	allnode_bits = * ((unsigned int *) MGT_ALLNODEID);
	unsigned int	*nodes = (unsigned int *) MASTER_LINES_NODE;
	unsigned int	*xs = (unsigned int *) MASTER_LINES_X;

	for (int i = 0; i <= y; i++) {
		if (! xs[i] && nodes[i] && slave_retrieve_lines_x(nodes[i]))
	 		UpdateLinesData (i);
	}

	return ;
}

void SlaveCheckLinesDone (int y, unsigned int *lines_x) 
{
	unsigned int	*xs = (unsigned int *) MASTER_LINES_X;
	unsigned char *mp = (unsigned char *) MASTER_LINES_POINT;

	for (int i = 0; i < 200; i++) {
		if ((xs[i] == 64) && (lines_x[i] != xs[i])) {
//printf ("Slave: lines_x[%d] set to %d\n", i, xs[i]);
			lines_x[i] = xs[i];
		}
	}

	return ;
}

void clearNodes ()
{
	unsigned int *nodes = (unsigned int *)MASTER_LINES_NODE;
	unsigned int *xs = (unsigned int *)MASTER_LINES_X;

	memset (nodes, 0, sizeof (unsigned int) * 200);
	memset (xs, 0, sizeof (unsigned int) * 200);
#if 0
	CommitChanges (nodes, sizeof (unsigned int)*200);
	CommitChanges (xs, sizeof (unsigned int)*200);
#endif
	CommitChanges ((void *) MASTER_NTIME, 34400);
}

#endif
